@@ -2,7 +2,7 @@
 // It is subject to the license terms in the LICENSE file found in the top-level directory
 // of this distribution and at http://opencv.org/license.html.
 //
-// Copyright (C) 2020 Intel Corporation
+// Copyright (C) 2020-2021 Intel Corporation
 
 #include <set> // set
 #include <map> // map
@@ -32,6 +32,14 @@ void putData(GSerialized& s, const cv::gimpl::GModel::ConstGraph& cg, const ade:
         });
     if (s.m_datas.end() == it) {
         s.m_datas.push_back(gdata);
+
+        if (cg.metadata(nh).contains<gimpl::ConstValue>()) {
+            size_t datas_num = s.m_datas.size() - 1;
+            GAPI_DbgAssert(datas_num <= static_cast<size_t>(std::numeric_limits<GSerialized::data_tag_t>::max()));
+            GSerialized::data_tag_t tag = static_cast<GSerialized::data_tag_t>(datas_num);
+            s.m_const_datas.emplace(tag,
+                                    cg.metadata(nh).get<gimpl::ConstValue>());
+        }
     }
 }
 
@@ -42,11 +50,20 @@ void putOp(GSerialized& s, const cv::gimpl::GModel::ConstGraph& cg, const ade::N
     s.m_ops.push_back(op);
 }
 
-void mkDataNode(ade::Graph& g, const cv::gimpl::Data& data) {
+ade::NodeHandle mkDataNode(ade::Graph& g, const cv::gimpl::Data& data) {
     cv::gimpl::GModel::Graph gm(g);
     auto nh = gm.createNode();
     gm.metadata(nh).set(cv::gimpl::NodeType{cv::gimpl::NodeType::DATA});
     gm.metadata(nh).set(data);
+    return nh;
+}
+
+ade::NodeHandle mkConstDataNode(ade::Graph& g, const cv::gimpl::Data& data, const cv::gimpl::ConstValue& const_data) {
+    auto nh = mkDataNode(g, data);
+
+    cv::gimpl::GModel::Graph gm(g);
+    gm.metadata(nh).set(const_data);
+    return nh;
 }
 
 void mkOpNode(ade::Graph& g, const cv::gimpl::Op& op) {
@@ -71,7 +88,7 @@ void linkNodes(ade::Graph& g) {
     for (const auto& nh : g.nodes()) {
         if (gm.metadata(nh).get<cv::gimpl::NodeType>().t == cv::gimpl::NodeType::OP) {
             const auto& op = gm.metadata(nh).get<gimpl::Op>();
-            for (const auto& in : ade::util::indexed(op.args)) {
+            for (const auto in : ade::util::indexed(op.args)) {
                 const auto& arg = ade::util::value(in);
                 if (arg.kind == cv::detail::ArgKind::GOBJREF) {
                     const auto idx = ade::util::index(in);
@@ -82,9 +99,9 @@ void linkNodes(ade::Graph& g) {
                 }
             }
 
-            for (const auto& out : ade::util::indexed(op.outs)) {
-                const auto idx = ade::util::index(out);
-                const auto rc  = ade::util::value(out);
+            for (const auto out : ade::util::indexed(op.outs)) {
+                const auto  idx = ade::util::index(out);
+                const auto& rc  = ade::util::value(out);
                 const auto& out_nh = dataNodes.at(rc);
                 const auto& out_eh = g.link(nh, out_nh);
                 gm.metadata(out_eh).set(cv::gimpl::Output{idx});
@@ -159,6 +176,13 @@ IIStream& operator>> (IIStream& is, cv::Point2f& pt) {
     return is >> pt.x >> pt.y;
 }
 
+IOStream& operator<< (IOStream& os, const cv::Point3f &pt) {
+    return os << pt.x << pt.y << pt.z;
+}
+IIStream& operator>> (IIStream& is, cv::Point3f& pt) {
+    return is >> pt.x >> pt.y >> pt.z;
+}
+
 IOStream& operator<< (IOStream& os, const cv::Size &sz) {
     return os << sz.width << sz.height;
 }
@@ -184,18 +208,20 @@ IOStream& operator<< (IOStream& os, const cv::RMat& mat) {
     return os;
 }
 IIStream& operator>> (IIStream& is, cv::RMat&) {
-    util::throw_error(std::logic_error("operator>> for RMat should never be called"));
+    util::throw_error(std::logic_error("operator>> for RMat should never be called. "
+                                        "Instead, cv::gapi::deserialize<cv::GRunArgs, AdapterTypes...>() "
+                                        "should be used"));
     return is;
 }
 
-IOStream& operator<< (IOStream& os, const cv::MediaFrame &) {
-    // Stub
-    GAPI_Assert(false && "cv::MediaFrame serialization is not supported!");
+IOStream& operator<< (IOStream& os, const cv::MediaFrame &frame) {
+    frame.serialize(os);
     return os;
 }
 IIStream& operator>> (IIStream& is, cv::MediaFrame &) {
-    // Stub
-    GAPI_Assert(false && "cv::MediaFrame serialization is not supported!");
+    util::throw_error(std::logic_error("operator>> for MediaFrame should never be called. "
+                                        "Instead, cv::gapi::deserialize<cv::GRunArgs, AdapterTypes...>() "
+                                        "should be used"));
     return is;
 }
 
@@ -264,7 +290,7 @@ IOStream& operator<< (IOStream& os, const cv::Mat &m) {
     case CV_32S: write_mat_data< int32_t>(os, m); break;
     case CV_32F: write_mat_data<   float>(os, m); break;
     case CV_64F: write_mat_data<  double>(os, m); break;
-    default: GAPI_Assert(false && "Unsupported Mat depth");
+    default: GAPI_Error("Unsupported Mat depth");
     }
     return os;
 }
@@ -280,7 +306,7 @@ IIStream& operator>> (IIStream& is, cv::Mat& m) {
     case CV_32S: read_mat_data< int32_t>(is, m); break;
     case CV_32F: read_mat_data<   float>(is, m); break;
     case CV_64F: read_mat_data<  double>(is, m); break;
-    default: GAPI_Assert(false && "Unsupported Mat depth");
+    default: GAPI_Error("Unsupported Mat depth");
     }
     return is;
 }
@@ -293,10 +319,10 @@ IIStream& operator>> (IIStream& is,       cv::gapi::wip::draw::Text &t) {
 }
 
 IOStream& operator<< (IOStream&, const cv::gapi::wip::draw::FText &) {
-    GAPI_Assert(false && "Serialization: Unsupported << for FText");
+    GAPI_Error("Serialization: Unsupported << for FText");
 }
 IIStream& operator>> (IIStream&,       cv::gapi::wip::draw::FText &) {
-    GAPI_Assert(false && "Serialization: Unsupported >> for FText");
+    GAPI_Error("Serialization: Unsupported >> for FText");
 }
 
 IOStream& operator<< (IOStream& os, const cv::gapi::wip::draw::Circle &c) {
@@ -372,19 +398,19 @@ IIStream& operator>> (IIStream& is,       cv::GArrayDesc &) {return is;}
 #if !defined(GAPI_STANDALONE)
 IOStream& operator<< (IOStream& os, const cv::UMat &)
 {
-    GAPI_Assert(false && "Serialization: Unsupported << for UMat");
+    GAPI_Error("Serialization: Unsupported << for UMat");
     return os;
 }
 IIStream& operator >> (IIStream& is, cv::UMat &)
 {
-    GAPI_Assert(false && "Serialization: Unsupported >> for UMat");
+    GAPI_Error("Serialization: Unsupported >> for UMat");
     return is;
 }
 #endif // !defined(GAPI_STANDALONE)
 
 IOStream& operator<< (IOStream& os, const cv::gapi::wip::IStreamSource::Ptr &)
 {
-    GAPI_Assert(false && "Serialization: Unsupported << for IStreamSource::Ptr");
+    GAPI_Error("Serialization: Unsupported << for IStreamSource::Ptr");
     return os;
 }
 IIStream& operator >> (IIStream& is, cv::gapi::wip::IStreamSource::Ptr &)
@@ -403,7 +429,7 @@ struct putToStream<Ref, std::tuple<>>
 {
     static void put(IOStream&, const Ref &)
     {
-        GAPI_Assert(false && "Unsupported type for GArray/GOpaque serialization");
+        GAPI_Error("Unsupported type for GArray/GOpaque serialization");
     }
 };
 
@@ -428,7 +454,7 @@ struct getFromStream<Ref, std::tuple<>>
 {
     static void get(IIStream&, Ref &, cv::detail::OpaqueKind)
     {
-        GAPI_Assert(false && "Unsupported type for GArray/GOpaque deserialization");
+        GAPI_Error("Unsupported type for GArray/GOpaque deserialization");
     }
 };
 
@@ -534,7 +560,7 @@ IOStream& operator<< (IOStream& os, const cv::GArg &arg) {
         case cv::detail::OpaqueKind::CV_RECT:    os << arg.get<cv::Rect>();     break;
         case cv::detail::OpaqueKind::CV_SCALAR:  os << arg.get<cv::Scalar>();   break;
         case cv::detail::OpaqueKind::CV_MAT:     os << arg.get<cv::Mat>();      break;
-        default: GAPI_Assert(false && "GArg: Unsupported (unknown?) opaque value type");
+        default: GAPI_Error("GArg: Unsupported (unknown?) opaque value type");
         }
     }
     return os;
@@ -565,12 +591,13 @@ IIStream& operator>> (IIStream& is, cv::GArg &arg) {
             HANDLE_CASE(STRING  , std::string);
             HANDLE_CASE(POINT   , cv::Point);
             HANDLE_CASE(POINT2F , cv::Point2f);
+            HANDLE_CASE(POINT3F , cv::Point3f);
             HANDLE_CASE(SIZE    , cv::Size);
             HANDLE_CASE(RECT    , cv::Rect);
             HANDLE_CASE(SCALAR  , cv::Scalar);
             HANDLE_CASE(MAT     , cv::Mat);
 #undef HANDLE_CASE
-        default: GAPI_Assert(false && "GArg: Unsupported (unknown?) opaque value type");
+        default: GAPI_Error("GArg: Unsupported (unknown?) opaque value type");
         }
     }
     return is;
@@ -624,6 +651,10 @@ IOStream& operator<< (IOStream& os, const cv::gimpl::Data &d) {
     return os << d.shape << d.rc << d.meta << d.storage << d.kind;
 }
 
+IOStream& operator<< (IOStream& os, const cv::gimpl::ConstValue &cd) {
+    return os << cd.arg;
+}
+
 namespace
 {
 template<typename Ref, typename T, typename... Ts>
@@ -634,7 +665,7 @@ struct initCtor<Ref, std::tuple<>>
 {
     static void init(cv::gimpl::Data&)
     {
-        GAPI_Assert(false && "Unsupported type for GArray/GOpaque deserialization");
+        GAPI_Error("Unsupported type for GArray/GOpaque deserialization");
     }
 };
 
@@ -667,6 +698,9 @@ IIStream& operator>> (IIStream& is, cv::gimpl::Data &d) {
     return is;
 }
 
+IIStream& operator>> (IIStream& is, cv::gimpl::ConstValue &cd) {
+    return is >> cd.arg;
+}
 
 IOStream& operator<< (IOStream& os, const cv::gimpl::DataObjectCounter &c) {
     return os << c.m_next_data_id;
@@ -709,18 +743,34 @@ void serialize( IOStream& os
     }
     s.m_counter = cg.metadata().get<cv::gimpl::DataObjectCounter>();
     s.m_proto   = p;
-    os << s.m_ops << s.m_datas << s.m_counter << s.m_proto;
+    os << s.m_ops << s.m_datas << s.m_counter << s.m_proto << s.m_const_datas;
 }
 
 GSerialized deserialize(IIStream &is) {
     GSerialized s;
-    is >> s.m_ops >> s.m_datas >> s.m_counter >> s.m_proto;
+    is >> s.m_ops >> s.m_datas >> s.m_counter >> s.m_proto >> s.m_const_datas;
     return s;
 }
 
 void reconstruct(const GSerialized &s, ade::Graph &g) {
     GAPI_Assert(g.nodes().empty());
-    for (const auto& d  : s.m_datas) cv::gapi::s11n::mkDataNode(g, d);
+
+    GSerialized::data_tag_t tag = 0;
+    for (const auto& d  : s.m_datas) {
+        if (d.storage == gimpl::Data::Storage::CONST_VAL) {
+            auto cit = s.m_const_datas.find(tag);
+            if (cit == s.m_const_datas.end()) {
+                util::throw_error(std::logic_error("Cannot reconstruct graph: Data::Storage::CONST_VAL by tag: " +
+                                  std::to_string(tag) + " requires ConstValue"));
+            }
+
+            mkConstDataNode(g, d, cit->second);
+        } else {
+            cv::gapi::s11n::mkDataNode(g, d);
+        }
+
+        tag ++;
+    }
     for (const auto& op : s.m_ops)   cv::gapi::s11n::mkOpNode(g, op);
     cv::gapi::s11n::linkNodes(g);
 
@@ -886,7 +936,7 @@ IIStream& ByteMemoryInStream::operator>> (std::string& str) {
     if (sz == 0u) {
         str.clear();
     } else {
-        str.resize(sz);
+        str.resize(static_cast<std::size_t>(sz));
         for (auto &&i : ade::util::iota(sz)) { *this >> str[i]; }
     }
     return *this;
@@ -906,6 +956,9 @@ GAPI_EXPORTS void serialize(IOStream& os, const cv::GMetaArgs &ma) {
 GAPI_EXPORTS void serialize(IOStream& os, const cv::GRunArgs &ra) {
     os << ra;
 }
+GAPI_EXPORTS void serialize(IOStream& os, const std::vector<std::string> &vs) {
+    os << vs;
+}
 GAPI_EXPORTS GMetaArgs meta_args_deserialize(IIStream& is) {
     GMetaArgs s;
     is >> s;
@@ -916,6 +969,11 @@ GAPI_EXPORTS GRunArgs run_args_deserialize(IIStream& is) {
     is >> s;
     return s;
 }
+GAPI_EXPORTS std::vector<std::string> vector_of_strings_deserialize(IIStream& is) {
+    std::vector<std::string> s;
+    is >> s;
+    return s;
+}
 
 } // namespace s11n
 } // namespace gapi
